/*
Jack Midi Logger - See what MIDI message comes in.
    Copyright (C) 2018  Arnaud Estublier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see https://www.gnu.org/licenses/gpl.txt.
*/

#include "audio.hpp"

std::string audio::midi_notes::getNoteName(int noteNumber) {
    if( noteNumber > 127 || noteNumber < 0 ) {
        return std::string("Undefined");
    } else {
        std::string res;
        res += m_noteNames[noteNumber%12];
        res += std::to_string(std::div(noteNumber, 12).quot - 1);
        return res;
    }
}

std::string audio::midi_control_change::getControlByNumber(int ccNumber) {
    if( ccNumber > 127 || ccNumber < 0 ) {
        return std::string("Undefined");
    } else {
        return this->m_controlChange[ccNumber];
    }
}

audio::midi_input_port::midi_input_port(std::string name, std::string source, unsigned int index, jack_client_t* client ) {
    m_index = index;
    m_source = source;
    m_name = name;
    m_client = client;

    m_port = jack_port_register(m_client, m_name.c_str(), JACK_DEFAULT_MIDI_TYPE, JackPortIsInput, 0);

    if ( m_port) {
        std::cout << "jack port: " << m_name << " created" << std::endl;
        if( !jack_connect(m_client, m_source.c_str(), jack_port_name(m_port)) ) {
            m_con_state = true;
            std::cout << "jack port: " << m_name << " connected to ";
            std::cout << m_source << std::endl;
        } else {
            m_jack_errors++;
            m_con_state = false;
            std::cout << "jack port: " << name << " not connected to ";
            std::cout << m_source << std::endl;
        }
    } else {
        m_jack_errors++;
        std::cout << "jack port: " << m_name << " not created ?" << std::endl;
    }
}

void audio::midi_input_port::disconnect() {
    if( !this->m_jack_errors ) {
        if( jack_port_disconnect(this->m_client, this->m_port) ) {
            std::cerr << "jack port: " << this->m_name << " not disconnected" << std::endl;
            this->m_jack_errors++;
        } else {
            std::cout << "jack port: " << this->m_name << " disconnected" << std::endl;
        }
    } else {
        std::cout << "jack port: " << this->m_name << " was not created, going out without processing..." << std::endl;
    }

}
void audio::midi_input_port::unregister() {
    if( !this->m_jack_errors ) {
        if( jack_port_unregister(this->m_client, this->m_port) ) {
            std::cerr << "jack port: " << this->m_name << " not unregistered" << std::endl;
            this->m_jack_errors++;
        } else {
            std::cout << "jack port: " << this->m_name << " unregistered" << std::endl;
        }
    } else {
        std::cout << "jack port: " << this->m_name << " was not created, going out without processing..." << std::endl;
    }
}

std::string audio::midi_input_port::name() {return this->m_name;};
std::string audio::midi_input_port::source() {return this->m_source;};
jack_port_t* audio::midi_input_port::port_ptr() {return this->m_port;};
unsigned int audio::midi_input_port::index() {return this->m_index;};
void audio::midi_input_port::reconnect(std::string name, std::string source, unsigned int index) {
    bool hasChanged = false;

    if( name != this->m_name ) {
        this->m_name = name;
        jack_port_set_name(this->m_port, this->m_name.c_str());
        hasChanged = true;
        std::cout << "jack port: updated name to " << this->m_name << std::endl;
    }
    if( source != this->m_source ) {
        this->m_source = source;
        hasChanged = true;
        std::cout << "jack port: updated input port to " << this->m_source << std::endl;
    }
    if ( index != this->m_index  ) {
        this->m_index = index;
        hasChanged = true;
    }

    if( !jack_port_connected(this->m_port) || hasChanged ) {
        jack_port_disconnect(this->m_client, this->m_port);
        jack_connect(this->m_client, this->m_source.c_str(), jack_port_name(this->m_port));
    }
}

bool audio::midi_input_port::is_safe() {
    return ( this->m_jack_errors ) ? false : true;
}

audio::midi_client::~midi_client() {
    bool unregister = false;

    if(!this->m_ports.empty()) {
        for(std::vector<audio::midi_input_port>::iterator port=this->m_ports.begin();port!=this->m_ports.end(); port++) {
            port->disconnect();
            port->unregister();
            unregister = unregister && port->is_safe();
        }
    }
    this->m_ports.clear();

    if(unregister) {
        std::cerr << "jack: ports not unregistered" << std::endl;
    } else {
        std::cout << "jack: ports unregistered" << std::endl;
    }

    if( jack_deactivate(this->m_client) ) {;
        std::cerr << "jack: client not deactivated" << std::endl;
    }else {
        std::cout << "jack: client deactivated" << std::endl;
    };
    if( jack_client_close(this->m_client) ) {;
        std::cerr << "jack: client not closed" << std::endl;
    }else {
        std::cout << "jack: client closed" << std::endl;
    };
}

void audio::midi_client::activate() {
    if( (this->m_client = jack_client_open(this->m_name, JackNoStartServer, NULL)) ) {
        std::cout << "jack: client opened" << std::endl;
        if( jack_set_port_registration_callback(this->m_client, cb_registration, this) || jack_set_process_callback(this->m_client, cb_process, this) ) {
            this->m_jack_errors++;
            std::cerr << "jack: callback error" << std::endl;
        } else {
            std::cout << "jack: callback attached" << std::endl;
        };
    } else {
        this->m_jack_errors++;
        std::cerr << "jack: client not open" << std::endl;
    }

    if( jack_activate(this->m_client) ) {
        this->m_jack_errors++;
        std::cerr << "jack: client not activated" << std::endl;
    }else {
        std::cout << "jack: client activated" << std::endl;
        this->init_midi_inputs();
    };
}

bool audio::midi_client::isActivated() {
    return ( this->m_jack_errors > 0 ) ? false : true;
}

void audio::midi_client::init_midi_inputs() {
    const char ** midi_in_clients = jack_get_ports(this->m_client, NULL, "midi", JackPortIsOutput);
    std::vector<std::string> res;

    if(midi_in_clients) {
        std::cout << "jack: init registration" << std::endl;
        for(int i=0; midi_in_clients[i]; i++) {
            res.push_back("1");
            res.push_back(std::string(midi_in_clients[i]));
            this->m_q_midiInputs.push(res);
            res.clear();
        }

        jack_free(midi_in_clients);
        std::cout << "jack: end init registration" << std::endl;
    }
}

void audio::midi_client::cb_registration(jack_port_id_t port, int regis, void *arg) {
    audio::midi_client * p = (audio::midi_client *) arg;
    jack_port_t * reg_port = jack_port_by_id(p->m_client, port);
    std::vector<std::string> res;
    if( !jack_port_is_mine(p->m_client, reg_port) && jack_port_flags(reg_port)%2 == 0) {
        if( !regis ) {
            res.push_back(std::to_string(regis));
            res.push_back(std::string(jack_port_name(reg_port)));
            p->m_q_midiInputs.push(res);
        } else if ( !std::strcmp(jack_port_type(reg_port), "8 bit raw midi")) {
            res.push_back(std::to_string(regis));
            res.push_back(std::string(jack_port_name(reg_port)));
            p->m_q_midiInputs.push(res);
        }
    }
};

int audio::midi_client::cb_process(jack_nframes_t nframes, void* arg) {
    audio::midi_client * c = (audio::midi_client *) arg;
    for(std::vector<audio::midi_input_port>::iterator port=c->m_ports.begin(); port!=c->m_ports.end(); port++) {
        void * port_buff = jack_port_get_buffer(port->port_ptr(), nframes);
        jack_midi_event_t in_event;
        jack_nframes_t event_count = jack_midi_get_event_count(port_buff);

        if(event_count) {
            //char * aliases[2];
            std::vector<unsigned char> message;
            //jack_port_get_aliases(*port, aliases);
            for(unsigned int i=0; i < event_count; i++) {
                jack_midi_event_get(&in_event, port_buff, i);
                //message.push_back(std::atoi(aliases[0]));
                message.push_back(port->index());
                message.push_back(in_event.size);
                for(long unsigned int j=0; j<in_event.size; j++) {
                    message.push_back(*(in_event.buffer+j));
                }
                c->m_q_midiMessages.push(message);
                message.clear();
            }
        }
    }

    return 0;
}

void audio::midi_client::check_port(std::vector<std::string>& ports) {
    std::string name;
    std::string compare;
    bool exists = false;
    name.append("midi_in_");
    if( ports[0].size() == 1 ) {
        ports[0].insert(0, std::to_string(0));
    }
    name.append(ports[0]);
    compare.append(this->m_name);
    compare.append(":");
    compare.append(name);

    if(!this->m_ports.empty()) {
        for(std::vector<audio::midi_input_port>::iterator port=this->m_ports.begin();port!=this->m_ports.end(); port++) {
            if( port->name() == name || port->source() == ports[2]) {
                exists = true;
                if( (ports[1] == "1") ) {
                    port->reconnect(name, ports[2], std::stoi(ports[0], nullptr, 10));
                } else if ( ports[1] == "0" ) {
                    port->disconnect();
                    port->unregister();
                    this->m_ports.erase(port);
                }
                this->m_ports.shrink_to_fit();
                break;
            }
        }
    }
    if(!exists && ports[1] == "1") {
        this->m_ports.push_back(audio::midi_input_port(name, ports[2], std::stoi(ports[0], nullptr, 10), this->m_client));
    };
}
